# OBJ - timestamped JSON objects
#
# obj/base.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 18-09-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

"""
    providing timestamped JSON objects and their load/save methods.
    
"""

__txt__ = "Timestamped JSON Objects"
__version__ = 83
                 
## imports

import _thread
import collections
import datetime
import logging
import os
import importlib
import json
import sys
import time
import types
import threading

## defines

starttime = time.time()

## exceptions

class EKEY(Exception):

    """ value is a function or method and is already present and cannot be overriden. """

    pass


class ENOTPATH(Exception):

    """ path is missing or not correct. """
    
    pass

## classes

class Object(object):

    """ The pure object that does load/save from and to JSON files. """

    def __repr__(self) -> str:
        return "%s.%s at %s" % (
            self.__class__.__module__,
            self.__class__.__name__,
            str(hex(id(self)))
        )

    def __str__(self) -> str:
        return json.dumps(self, default=smooth, indent=4, sort_keys=True)

    def get(self, k, d=""):
        """ get attribute from object. """
        return getattr(self, k, d)

    def load(self, path: str) -> None:
        """ load an object from given path. """
        assert path
        p = os.path.abspath(os.path.join(cfg.workdir, path)) 
        logging.debug("load %s" % path)
        with open(p) as f:
            data = json.load(f)
            for k, v in data.items():
                self.set(k ,v)
        if "_container" not in dir(self):
            self._container = {}
        self._container["path"] = path
        return self

    def match(self, m: dict) -> bool:
        """ return whether this object values matches criteria. """ 
        res = False 
        for k in m.keys():
            if m[k] == "" and self.get(k, ""):
                res = True
            elif m[k] in self.get(k, ""):
                res = True
            else:
                res = False
                break
        return res

    def exclusive(self, m: dict) -> bool:
        """ return whether this object values matches criteria. """ 
        res = False 
        for k in m.keys():
            if m[k] and m[k] in self.get(k, ""):
                res = True
            else:
                res = False
                break
        return res

    def save(self, path="", stime="") -> str:
        """ save this object to a file. """
        if not path and not stime:
            try:
                path = self._container["path"]
            except (TypeError, AttributeError, KeyError):
                pass
        if not path:
            path = os.path.join(get_type(self), stime or rtime())
        if path:
            if "_container" not in dir(self):
                self._container = {}
                self._container["path"] = path
        if not path:
            raise ENOPATH
        logging.info("save %s" % path)
        p = os.path.abspath(os.path.join(cfg.workdir, path)) 
        d, fn = os.path.split(p)
        if not os.path.isdir(d):
            cdir(d)
        with open(p, "w") as f:
            json.dump(self, f, default=smooth, indent=4, sort_keys=True)
        return path

    def set(self, k , v):
        """ set a attribute. """
        return setattr(self, k, v)

class Dict(Object, dict):

    """ Object (load/save) mixin with a dict. """

    def upgrade(self, d):
        for k, v in d.items():
            if v:
                self[k] = v
        return self

class Dotted(Dict):

    """ JSON load/saveable dict with dotted access (like a object). """

    def __getattribute__(self, name):
        try:
            val = super().__getattribute__(name)
        except AttributeError:
            try:
                val = self[name]
            except KeyError as ex:
                raise AttributeError(name)
        return val

    def __getattr__(self, name):
        if name not in self:
            if "_default" in self:
                self[name] = self["_default"]
            else:
                raise AttributeError(name)
        return self.__getitem__(name)

    def __dir__(self):
        return self.keys()

    def __setattr__(self, name, value):
        return self.__setitem__(name, value)

class Register(Dict):

    """ dict of list of values. """

    def register(self, k: str, v: object) -> None:
        """ register a key, value pair, constructing a list if key is not in object. """
        if k not in self:
            self[k] = []
        return self[k]

class Store(Dotted):

    """ manages objects stored on disk. """

    def get_fns(self, match=""):
        res = []
        for fn in filenames(match):
            res.append(fn)
        return sorted(res, key=lambda x: fn_time(x))

    def deleted(self, match):
        """ show objects with the deleted flag set. """
        for fn in self.get_fns(match):
            cls = get_cls(get_fntype(fn))
            o = cls().load(fn)
            if "_deleted" in dir(o) and o._deleted:
                yield o
        
    def find(self, match={}, type="", skip=True, strict=False):
        """ find matching objects. """
        if not match:
            return []
        for fn in self.get_fns(type):
            cls = get_cls(get_fntype(fn))
            o = cls().load(fn)
            if skip and "_deleted" in dir(o) and o._deleted:
                continue
            if not match:
                return
            if strict and o.exclusive(match):
                yield o
            elif o.match(match):
                yield o

    def last(self, match=""):
        """ show last object. """
        fns = self.get_fns(match)
        if fns:
            fn = fns[-1]
            cls = get_cls(get_fntype(fn))
            o = cls().load(fn)
            return o

## instances

cfg = Dotted()
cfg.workdir = os.path.join(os.path.expanduser("~"), ".obj")

## utility functions

def cdir(path) -> bool:
    """ create directory. """
    res = ""
    for p in path.split(os.sep):
        res += "%s%s" % (p, os.sep)
        p = os.path.abspath(os.path.normpath(res))
        try:
            os.mkdir(p)
        except (IsADirectoryError, NotADirectoryError, FileExistsError):
            pass
    return True

def filenames(type=None):
    """ show file names of objects stored on path (index into the workdir). """
    res = []
    if type:
        p = os.path.join(cfg.workdir, type)
    else:
        p = cfg.workdir
    for rootdir, dirs, files in os.walk(p, topdown=True):
        if p not in rootdir:
            continue
        for fn in files:
            yield os.path.join(rootdir, fn).split(cfg.workdir)[-1][1:]

def fn_time(daystr: str) -> None:
    """ return the time determined from the filename of an object. """
    daystr = daystr.replace("_", ":")
    datestr = " ".join(daystr.split(os.sep)[-2:])
    datestr = datestr.split(".")[0]
    try:
        t = time.mktime(time.strptime(datestr, "%Y-%m-%d %H:%M:%S"))
    except ValueError:
        t = 0
    return t

def get_cls(name):
    """ return class from sys.modules. """
    name, c = name.rsplit(".", 1)
    mod = importlib.import_module(name)
    return getattr(mod, c)

def get_type(obj):
    """ return this type of object. """
    return obj.__class__.__module__ + "." + obj.__class__.__qualname__

def get_fntype(fn):
    return fn.split(os.sep)[0]

def prefixes():
    """ show list of subdirectories in the workdir. """
    return sorted(set([x.split(".")[-1].lower() for x in os.listdir(cfg.workdir)]))

def rtime():
    """ return filename base of current time. """ 
    return str(datetime.datetime.now()).replace(" ", os.sep)

def smooth(o):
    """ deal with objects that are not convertable to basic JSON types. """
    if isinstance(o, Object):
        return vars(o)
    elif type(o) not in [dict, list, str, int, float, bool, None]:
        return repr(o)
    return o
