# OBJ - timestamped JSON objects
#
# obj/handler.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 18-09-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" queued event handler. """

__version__ = 1

## imports

import obj.base
import obj.event
import obj.fleet
import obj.loader
import obj.task
import obj.utils

import _thread
import logging
import pkgutil
import queue
import threading
import types

## classes

class Handler(obj.loader.Loader):

    """ this handler handles events. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ready = threading.Event()
        self._stopped = False
        self._type = obj.utils.name(self)
        self.cbs = obj.base.Register()
        self.cc = "!"
        self.handlers = {}
        self.names = names
        self.queue = queue.Queue()
        self.threaded = False

    def callback(self, event):
        """ call callbacks registered for eventtype stored in event.cb. """
        for cb in self.get_cbs(event.cb):
            obj.task.launch(cb, event)

    def dispatch(self, event):
        """ parse the event and call corresponding command. """
        event.parse()
        func = self.get_cmd(event)
        if func:
            event.func = func
            func(event)
            if event.batch:
                event.show(self)
        event.ready()
        
    def event(self, txt="", origin="", options="", default={}):
        """ return an event with corresponding arguments set. """
        e = obj.event.Event(default)
        e.txt = txt
        e.cc = self.cc
        e.options = options
        e.origin = origin or "root@shell"
        e.orig = repr(self)
        return e

    def get_cbs(self, name):
        """ return callbacks registered. """
        return self.cbs.get(name, [])

    def get_cmd(self, event):
        """ return commands. """
        cmd = event.get_cmd()
        return self.get_handler(cmd)

    def get_event(self):
        """ return event from queue. """
        return self.queue.get()

    def get_handler(self, cmd):
        """ return handler registered. """
        val = self.handlers.get(cmd, None)
        if not val:
            modname = self.names.get(cmd, None)
            if modname:
                self.load_mod(modname)
            val = self.handlers.get(cmd, None)
        return val

    def put(self, event):
        """ pass event over to queue. """
        self.queue.put_nowait(event)

    def list(self, n):
        """ return list of modules in the obj package. """
        res = []
        try:
            mod = self.direct(n)
        except:
            logging.error(obj.utils.get_exception())
            return res
        if mod:
            if "__path__" in dir(mod):
                for nn in pkgutil.walk_packages(mod.__path__, n+'.'):
                    res.append(nn[1])
            elif "__file__" in dir(mod):
                for nn in pkgutil.walk_packages([mod.__file__,], n):
                    res.append(nn[1])
        return res

    def load_mod(self, modname, mod=None):
        """ load a module into the table and scan it for commands. """
        if not mod:
            mod = super().load_mod(modname)
        if not mod:
            return
        for key, func, modname in self.scan(mod):
            if "callback" in dir(func) and func.callback:
                self.cbs.register(func.callback, func)
            else:
                self.handlers[key] = func
                self.names[key] = modname
        return mod

    def loop(self):
        """ the event loop. """
        while not self._stopped:
            try:
                event = self.get_event()
            except (EOFError, KeyboardInterrupt):
                _thread.interrupt_main()
                return
            if not event:
                break
            if self.threaded or event.threaded:
                launch(self.dispatch, event)
            else:
                try:
                    self.dispatch(event)
                except Exception as ex:
                    logging.error("%s %s" % (event.txt, obj.utils.get_exception()))

    def ready(self):
        """ flag handler as ready. """
        self._ready.set()

    def register(self, cmd, callback):
        """ register a callback. """
        self.handlers[cmd] = callback

    def scan(self, mod):
        """ scan a module for commands. """
        for key in dir(mod):
            if key.startswith("_"):
                continue
            func = getattr(mod, key, None)
            if func and type(func) == types.FunctionType:
                if "event" in func.__code__.co_varnames:
                    yield (key, func, mod.__name__)

    def start(self):
        """ start the event loop and register this handler with the bus. """
        fleet = obj.fleet.Fleet()
        fleet.add(self)
        obj.task.launch(self.loop)
         
    def stop(self): 
        """ stop this handler. """
        self._stopped = True
        self.queue.put(None)

    def unload(self, modname):
        """ unload a module. """
        super().unload(modname)
        for key, func, name in self.scan(modname):
            if name == modname:
                del self[key]

    def wait(self):
        """ wait for handler to finish. """
        return self._ready.wait()

    def walk(self, name):
        """ browse a packages for all modules in it. """
        for n in self.list(name):
            if name.lower() in n.lower():
                mod = self.load_mod(n)
                if "start" in dir(mod):
                    mod.start()

names = obj.base.Dotted({
        "announce": "obj.cmds",
        "cfg": "obj.cmds",
        "cmds": "obj.cmds",
        "ed": "obj.cmds",
        "find": "obj.cmds",
        "fleet": "obj.cmds",
        "kill": "obj.cmds",
        "load": "obj.cmds",
        "log": "obj.cmds",
        "ls": "obj.cmds",
        "meet": "obj.cmds",
        "perm": "obj.cmds",
        "ps": "obj.cmds",
        "rm": "obj.cmds",
        "rmperm": "obj.cmds",
        "save": "obj.cmds",
        "test": "obj.cmds",
        "todo": "obj.cmds",
        "undel": "obj.cmds",
        "unload": "obj.cmds",
        "uptime": "obj.cmds",
        "user": "obj.cmds",
        "version": "obj.cmds"
})
