# OBJ - timestamped JSON objects
#
# obj/task.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 18-09-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

"""
    The task module provides one class, namely Task.
    A Task is the Thread version of OBJ that can be used to run code in a context.
    Tasks names are in the form of class.method.
    4 utility functions are provides to launch tasks, get a list of running tasks or kill a task.

    workflow:

    >>> from obj.event import Event
    >>> from obj.task import launch
    >>> def echo(event): event.reply(event.txt)
    >>> launch(hello, e)

"""

__version__ = 1

## imports

import obj
import queue
import threading
import time

## classes

class Task(threading.Thread):

    def __init__(self, *args, **kwargs):
        super().__init__(None, self.run, "", [], kwargs, daemon=False)
        self._queue = queue.Queue()
        self.result = None
        
    def __iter__(self):
        return self

    def __next__(self):
        for k in dir(self):
            yield k

    def put(self, *args, **kwargs):
        self._queue.put((args, kwargs))

    def run(self):
        import logging
        (args, kwargs) = self._queue.get()
        self.func = args[0]
        self.setName(obj.utils.name(self.func))
        try:
            self.result = self.func(*args[1:], **kwargs)
        except Exception as ex:
            logging.error(obj.utils.get_exception())

    def join(self):
        super().join()
        return self.result

## utility functions

def launch(*args, **kwargs):
    """ launch a tasks, use launch(function, arg1, arg2) etc. """
    t = Task()
    t.put(*args, **kwargs)
    t.start()
    return t

def kill(thrname=""):
    """ kill tasks, if thrname is provided an uppername match if done on the threadname. """
    for task in running(thrname):
        if "cancel" in dir(task):
            task.cancel()
        if "exit" in dir(task):
            task.exit()
        if "stop" in dir(task):
            task.stop()

def running(tname=""):
    """ show running tasks. """
    for task in threading.enumerate():
        name = str(task)
        if name.startswith("<_"):
            continue
        if tname and tname.upper() not in name.upper():
            continue
        yield task
